from typing import Dict, Optional
from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict, field
from promptrix.promptrixTypes import PromptMemory, Tokenizer, PromptFunctions
from alphawave_agents.agentTypes import TaskResponse
from alphawave_agents.SchemaBasedCommand import CommandSchema as sbcCommandSchema
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand


@dataclass
class CommandSchema(sbcCommandSchema):
    schema_type: str = "object"
    title: str = "finalAnswer"
    description: str = "generate an answer for the user"
    properties: dict = field(default_factory=lambda: {
        "answer": {
            "type": "string",
            "description": "final answer"
        }
    })
    required: list[str] = field(default_factory=lambda: ["answer"])
    returns: str = "a followup task or question"

@dataclass
class FinalAnswerCommandInput:
    answer:str

class FinalAnswerCommand(SchemaBasedCommand):
    def __init__(self, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__(CommandSchema(), title, description)

    def execute(self, input: FinalAnswerCommandInput, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer) -> TaskResponse:
        return {
            "type": "TaskResponse",
            "status": "success",
            "message": input['answer']
        }



from dataclasses import dataclass, asdict
from typing import Any, Dict, Optional, Union
from uuid import uuid4
from pyee import AsyncIOEventEmitter
import asyncio, copy
import json
import sys
import traceback
from promptrix.promptrixTypes import Message, PromptFunctions, PromptSection, PromptMemory, Tokenizer
from promptrix.FunctionRegistry import  FunctionRegistry
from promptrix.GroupSection import GroupSection
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.VolatileMemory import VolatileMemory
from promptrix.ConversationHistory import ConversationHistory
from promptrix.TextSection import TextSection
from promptrix.Utilities import Utilities
from promptrix.TemplateSection import TemplateSection
from promptrix.Prompt import Prompt

from alphawave.AlphaWave import AlphaWave, AlphaWaveOptions
from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse, Validation, PromptResponseValidator
from alphawave.JSONResponseValidator import JSONResponseValidator

from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand, CommandSchema
from alphawave_agents.AgentCommandValidator import AgentCommandValidator
import promptrix.Utilities
from alphawave_agents.agentTypes import Command, AgentThought, AgentThoughtSchema
from alphawave_agents.AgentCommandSection import AgentCommandSection

from typing import Dict, Any
from pydantic import BaseModel, Field

class AgentCommandSchema(BaseModel):
    type: str = Field("object", description="an agent that can perform a task")
    title: str = "Agent"
    properties: Dict[str, Any] = {
        "input": {
            "type": "string",
            "description": "input for command",
        }
    }
    required: list[str] = ["input"]

PromptInstructionSection = TemplateSection("\n".join([
    "Return a JSON object with your thoughts and the next command to perform",
    "Only respond with the JSON format below and based your plan on the commands above",
    "Response Format:",
    '{"thoughts":{"thought":"<your current thought>","reasoning":"<self reflect on why you made this decision>","plan":"- short bulleted\n- list that conveys\n- long-term plan"},"command":{"name":"<command name>","input":{"<name>":"<value>"}}}'
]), 'system')

@dataclass
class AgentOptions:
    client: PromptCompletionClient
    context_variable: Optional[str] = None
    prompt: Union[str, list[str], PromptSection] = None
    prompt_options: PromptCompletionOptions = None
    agent_variable: Optional[str] = None
    functions: Optional[PromptFunctions] = None
    history_variable: Optional[str] = None
    initial_thought: Optional[AgentThought] = None
    input_variable: Optional[str] = None
    logRepairs: Optional[bool] = None
    max_history_messages: Optional[int] = None
    max_repair_attempts: Optional[int] = None
    max_steps: Optional[int] = None
    memory: Optional[PromptMemory] = None
    retry_invalid_responses: Optional[bool] = None
    step_delay: Optional[int] = None
    tokenizer: Optional[Tokenizer] = None


@dataclass
class ConfiguredAgentOptions:
    agent_variable: str = None
    client: PromptCompletionClient = None
    context_variable: str = None
    functions: PromptFunctions = None
    history_variable: str = None
    initial_thought: Optional[AgentThought] = None
    input_variable: str = None
    logRepairs: bool = None
    max_history_messages: int = 10
    max_repair_attempts: int = 3
    max_steps: int = 5
    memory: PromptMemory = None
    prompt: Union[str, list[str], PromptSection] = None
    prompt_options: PromptCompletionOptions = None
    retry_invalid_responses: bool = False
    step_delay: int = 5
    tokenizer: Tokenizer = None

def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if key in instance and value is not None:
            instance[key] =  value


@dataclass
class AgentCommandInput:
    agentId: str
    input: str


@dataclass
class AgentState:
    totalSteps: int
    context: Optional[str] = None
    child: Optional[Dict[str, str]] = None


class Agent(SchemaBasedCommand):
    def __init__(self, options: AgentOptions, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__(CommandSchema, title, description)
        self._commands: Dict[str, Command] = {}
        self._options: ConfiguredAgentOptions = {
            'agent_variable': 'agent',
            'client': None,
            'context_variable': 'context',
            'functions': FunctionRegistry(),
            'history_variable': 'history',
            'initial_thought': None,
            'input_variable': 'input',
            'logRepairs': False,
            'max_history_messages': 10,
            'max_repair_attempts': 3,
            'max_steps': 5,
            'memory': VolatileMemory(),
            'prompt': None,
            'prompt_options': None,
            'retry_invalid_responses': False,
            'step_delay': 5,
            'tokenizer': GPT3Tokenizer(),
        }
        update_dataclass(self._options, **asdict(options))
        self._events: EventEmitter = AsyncIOEventEmitter()

    @property
    def client(self):
        return self._options.client

    @property
    def events(self):
        return self._events

    @property
    def functions(self):
        return self._options['functions']

    @property
    def memory(self):
        return self._options['memory']

    @property
    def options(self):
        return self._options

    @property
    def tokenizer(self):
        return self._options['tokenizer']

    # Command management

    def addCommand(self, command: Command):
        if command.title in self._commands:
            raise ValueError(f"A command with the title {command.schema.title} already exists.")
        self._commands[command.title] = command
        return self

    def getCommand(self, title: str):
        return self._commands.get(title)

    def hasCommand(self, title: str):
        return title in self._commands

    # Task execution

    async def completeTask(self, input: Optional[str] = None, agentId: Optional[str] = None, executeInitialThought: bool = False):
        # Initialize the input to the next step
        stepInput = input if input is not None else self.memory.get(self.options['input_variable'])
        # Dispatch to child agent if needed
        step = 0
        state = self.get_agent_state(agentId)
        if state.child:
            childAgent = self.getCommand(state.child.title)
            response = await childAgent.completeTask(input, state.child.agentId)
            if response.status != 'success':
                return response

            # Delete child and save state
            del state.child
            self.setAgentState(state, agentId)

            # Use agents response as input to the next step
            # We don't know how many steps the child agent took, so we'll just assume it took one
            stepInput = response.message
            step = 1
            executeInitialThought = False

        # Start main task loop
        while step < self.options['max_steps']:
            # Wait for step delay
            if step > 0 and self.options['step_delay'] > 0:
                await asyncio.sleep(self.options['step_delay'])

            # Execute next step
            result = await self.execute_next_step(stepInput, agentId, executeInitialThought)
            if isinstance(result, str):
                stepInput = result
            else:
                return result

            step += 1
            executeInitialThought = False

        # Return too many steps
        return {
            "type": "TaskResponse",
            "status": "too_many_steps",
            "message": "The current task has taken too many steps."
        }

    # Agent as Commands
    async def execute(self, input: AgentCommandInput, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer):
        # Initialize the agents state
        agentId = input.agentId
        state = self.get_agent_state(agentId)
        state['context'] = input.input
        self.set_agent_state(state, agentId)

        # Start the task
        return await self.completeTask(None, agentId)

    def get_agent_state(self, agentId: Optional[str] = None):
        key = f'{self.options.agent_variable}-{agentId}' if agentId else self._options['agent_variable']
        state = self.memory.get(key) or {}
        if 'totalSteps' not in state:
            state['totalSteps'] = 0
        return state

    def set_agent_state(self, state: AgentState, agentId: Optional[str] = None):
        key = f"{self.options['agent_variable']}-{agentId}" if agentId else self.options['agent_variable']
        self.memory.set(key, state)

    def get_agent_history_variable(self, agentId: Optional[str] = None):
        return f"{self._options['history_variable']}-{agentId}" if agentId else self._options['history_variable']


    async def execute_next_step(self, input: Optional[str] = None, agent_id: Optional[str] = None, execute_initial_thought: bool = False):
        try:
            state = self.get_agent_state(agent_id)

            # Create agents prompt section
            if isinstance(self._options['prompt'], list):
                agent_prompt = TemplateSection('\n'.join(self._options['prompt']), 'system')
            elif isinstance(self._options['prompt'], dict):
                agent_prompt = self._options['prompt']
            else:
                agent_prompt = TemplateSection(self._options['prompt'], 'system')
            
            # Ensure the context variable is set
            if 'context' in state:
                self.memory.set(self.options['context_variable'], state['context'])

            # Create prompt
            history_variable = self.get_agent_history_variable(agent_id)
            sections = [agent_prompt]
            sections.append(AgentCommandSection(self._commands))
            sections.append(PromptInstructionSection)
#            prompt = Prompt([
#                GroupSection(sections, 'system'),
#                ConversationHistory(history_variable, 1.0, True)
#            ])
            prompt = Prompt([
                agent_prompt, AgentCommandSection(self._commands), PromptInstructionSection,
                ConversationHistory(history_variable, 1.0, True)
            ])
            if input:
                prompt.sections.append(TextSection(input, 'user', -1, True, '\n', 'user: '))

                # Ensure input variable is set otherwise the history will be wrong.
                self.memory.set(self.options['input_variable'], input)


            if execute_initial_thought and self._options['initial_thought']:
                # Just use initial thought as response
                # - This is used when agents are being called as commands.
                response = {
                    'status': 'success',
                    'message': {'role': 'assistant', 'content': self._options['initial_thought']}
                }
            else:
                # Add initial thought to history
                if state['totalSteps'] == 0 and self._options['initial_thought']:
                    history = self.memory.get(history_variable) or []
                    history.append({'role': 'assistant', 'content': json.dumps(self._options['initial_thought'])})
                    self.memory.set(history_variable, history)

                # Create command validator
                validator = AgentCommandValidator(self._commands)

                # Create a wave for the prompt
                #wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.options.prompt_options, memory=memory, functions=functions, tokenizer=tokenizer)
                wave = AlphaWave(
                    client = self._options['client'],
                    prompt = prompt,
                    prompt_options = self._options['prompt_options'],
                    functions = self._options['functions'],
                    history_variable = history_variable,
                    input_variable = self._options['input_variable'],
                    max_history_messages = self._options['max_history_messages'],
                    max_repair_attempts = self._options['max_repair_attempts'],
                    memory = self._options['memory'],
                    tokenizer = self._options['tokenizer'],
                    logRepairs = self._options['logRepairs'],
                    validator = validator
                )

                # Complete the prompt
                max_attempts = 2 if self._options['retry_invalid_responses'] else 1
                for attempt in range(max_attempts):
                    response = await wave.completePrompt()
                    if response['status'] != 'invalid_response':
                        break

                # Ensure response succeeded
                if response['status'] != 'success':
                    return {
                        'type': "TaskResponse",
                        'status': response['status'],
                        'message': response['message']
                    }

            # Get agents thought and execute command
            message = response['message']
            thought = message['content']
            self._events.emit('newThought', thought)
            result = await self.execute_command(state, thought)

            # Check for task result and error
            task_response = result if isinstance(result, dict) and result.get('type') == 'TaskResponse' else None
            if task_response:
                if task_response['status'] in ['error', 'invalid_response', 'rate_limited', 'too_many_steps', 'too_long']:
                    return task_response

            # Update history
            history = self.memory.get(history_variable) or []
            if input:
                history.append({'role': 'user', 'content': input})
            history.append({'role': 'assistant', 'content': json.dumps(thought)})
            self.memory.set(history_variable, history)

            # Save the agents state
            state['totalSteps'] += 1
            self.set_agent_state(state, agent_id)

            # Return result
            return task_response if task_response else Utilities.to_string(self.tokenizer, result)
        except Exception as err:
            return {
                'type': "TaskResponse",
                'status': "error",
                'message': str(err)
            }

    async def execute_command(self, state: AgentState, thought: AgentThought):
        # Get command
        command_name = thought['command']['input']['name'] if thought['command']['name'] == 'character' else thought['command']['name']
        command = self._commands.get(command_name, None) or {}
        input = thought['command']['input'] or {}
        if isinstance(command, Agent):
            # Pass control to child agent
            agentId = str(uuid4())
            childAgent = command
            response = await childAgent.execute(input['input'], self.memory, self.functions, self.tokenizer)
            if response.status == 'success':
                # Just return the response message since agent completed without additional input
                return response.message
            elif response.status == 'input_needed':
                # Remember that we're talking to the agent
                state.child = {
                    'title': command_name,
                    'agentId': agentId
                }
                return response
            else:
                # Return the response since the agent failed
                return response
        else:
            # Execute command and return result
            response = command.execute(input, self.memory, self.functions, self.tokenizer)
            if 'coroutine' in str(type(response)).lower():
                return await response
            return response

    async def completeTask(self, input: Optional[str] = None, agentId: Optional[str] = None, executeInitialThought: bool = False):
        # Initialize the input to the next step
        stepInput = input if input is not None else self.memory.get(self.options['input_variable'])

        # Dispatch to child agent if needed
        step = 0
        state = self.get_agent_state(agentId)
        if 'child' in state and state['child'] is not None:
            childAgent = self.getCommand(state['child'].title)
            response = await childAgent.completeTask(input, state['child'].agentId)
            if response.status != 'success':
                return response

            # Delete child and save state
            del state.child
            self.setAgentState(state, agentId)

            # Use agents response as input to the next step
            # We don't know how many steps the child agent took, so we'll just assume it took one
            stepInput = response.message
            step = 1
            executeInitialThought = False

        # Start main task loop
        while step < self.options['max_steps']:
            # Wait for step delay
            
            if step > 0 and self.options['step_delay'] > 0:
                sys.stdout.flush()
                await asyncio.sleep(self.options['step_delay']/1000)

            # Execute next step
            result = await self.execute_next_step(stepInput, agentId, executeInitialThought)
            if isinstance(result, str):
                stepInput = result
            else:
                return result

            step += 1
            executeInitialThought = False

        # Return too many steps
        return {
            "type": "TaskResponse",
            "status": "too_many_steps",
            "message": "The current task has taken too many steps."
        }



import openai
import requests
import time
import traceback
import json
import os
from typing import Any, Dict, List, Tuple
import alphawave.LLMClient as client

openai.api_key = os.getenv("OPENAI_API_KEY")

GPT_35 = 'gpt-3.5-turbo'
GPT_4 = 'gpt-4'
FC_T5 = 'fs-t5-3b-v1.0'
VICUNA_7B = 'vicuna-7b'
VICUNA_13B = 'vicuna-13b'
WIZARD_13B = 'wizard-13b'
WIZARD_30B = 'wizard-30b'
GUANACO_33B = 'guanaco-33b'
MODEL = WIZARD_30B
#MODEL = GPT_35
#MODEL = GPT_4
#MODEL = GUANACO_33B

if not MODEL.startswith('gpt') and MODEL != WIZARD_30B and MODEL != GUANACO_33B:
  # this for connecting to fastchat server
  client.initialize(MODEL)

def ask_LLM(model, gpt_message, max_tokens=100, temp=0.7, top_p=1.0, tkroot = None, tkdisplay=None):
    completion = None
    response = ''
    try:
      if model == FC_T5 or model == VICUNA_7B or model == VICUNA_13B or model == WIZARD_13B:
        print(f'calling {model} via fastchat')
        completion = client.send_w_stream(messages=gpt_message, max_tokens = max_tokens, temperature=temp )
        if completion is None:
          print('None from '+model)
        else:
          #print(f'completion from t5: {completion.choices[0].message.content}')
          response = completion

      elif model == WIZARD_30B or model == GUANACO_33B:
        print('calling {model} via local server')
        completion = client.run_query(gpt_message, temp, top_p, max_tokens, tkroot, tkdisplay)
        if completion is None:
          print('None from '+model)
        else:
          #print(f'completion from t5: {completion.choices[0].message.content}')
          if tkdisplay is not None:
            tkdisplay.insert(tk.END, completion)
            if tkroot is not None:
              tkroot.update()
          response = completion

      else:
        print('calling chatGPT', gpt_message)
        stream= openai.ChatCompletion.create(
            model=model, messages=gpt_message, max_tokens=max_tokens, temperature=temp, top_p=1, stop='STOP', stream=True)
        response = ''
        if stream is None:
          print('None from GPT')
          return response
        for chunk in stream:
          item = chunk['choices'][0]['delta']
          if 'content' in item.keys():
            #print(f"completion from gpt: {chunk['choices'][0]['delta']['content']}")
            if tkdisplay is not None:
              tkdisplay.insert(tk.END, chunk['choices'][0]['delta']['content'])
              if tkroot is not None:
                tkroot.update()
              response += chunk['choices'][0]['delta']['content']
    except:
        traceback.print_exc()
        print(response)
    return response

if __name__ == '__main__':  
  print(ask_LLM(ut.MODEL, gpt_message, max_tokens=100, temp=.1, top_p=0.1))



import argparse
import json
import traceback
import socket
import requests
from fastchat.conversation import Conversation, SeparatorStyle
import fastchat.conversation
import tkinter as tk
import time

MODEL_NAME = None; WORKER_ADDR=None; CONTROLLER_ADDRESS = "http://localhost:21001"
USER = '\n### HUMAN:\n'
ASSISTANT = '\n### ASSISTANT:\n'

host='192.168.1.195'
port = 5004
def run_query(messages, temp, top_p, max_tokens, tkroot = None, tkdisplay=None): 
    prompt = ''
    for msg in messages:
        if not isinstance(msg, dict):
            msg = msg.__dict__
        role = msg['role']
        if role.lower() == 'user' or role.lower() == 'system':
            role_os = USER
        elif role.lower() == 'assistant':
            role_os = ASSISTANT
        else: print(f'***** unknown role {role}')
        prompt += role_os + str(msg['content'])
    prompt += ASSISTANT
    server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens}
    smj = json.dumps(server_message)
    try:
        client_socket = socket.socket()  # instantiate
        client_socket.connect((host, port))  # connect to the server
        client_socket.settimeout(240)
        server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens}
        smj = json.dumps(server_message)
        client_socket.sendall(smj.encode('utf-8'))
        client_socket.sendall(b'x00xff')
        response = ''
        while True:
            s = client_socket.recv(1024) # break every 32 chars to stream output
            if s is None or not s:
                break
            if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                s = s[:-6].decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
                break
            else:
                s = s.decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
        client_socket.close()  # close the connection
        return response
    except: 
        traceback.print_exc()

    response = ''
    """try:
        client_socket = socket.socket()  # instantiate
        client_socket.connect((host, port))  # connect to the server
        client_socket.settimeout(240)
        
        client_socket.sendall(smj.encode('utf-8'))
        client_socket.sendall(b'x00xff')
        try:
            while True:
                s = client_socket.recv(64)
                if s is None or not s:
                    break
                if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                    s = s[:-6].decode('utf-8')
                    response +=  s
                    break
                else:
                    response +=  s.decode('utf-8')
            client_socket.close()  # close the connection
        except KeyboardInterrupt:
            pass
    except Exception:
        traceback.print_exc()
    return response
    """
    
def send(messages, temperature=0.0, max_tokens= 500):
    global MODEL_NAME, WORKER_ADDR, CONTROLLER_ADDRESS
    ###
    ### build prompt
    #
    conv=fastchat.conversation.get_conv_template('wizard')
    for msg in messages:
        role = msg['role']
        if role.lower() == 'user':
            role_index = 0
        else:
            role_index = 1
        conv.append_message(conv.roles[role_index], msg['content'])
    conv.append_message(conv.roles[1], '')
    prompt = conv.get_prompt()
    headers = {"User-Agent": "FastChat Client"}
    gen_params = {
        "model": MODEL_NAME,
        "prompt": prompt,
        "temperature": temperature,
        "max_new_tokens": max_tokens,
        "stop": conv.stop_str,
        "stop_token_ids": conv.stop_token_ids,
        "echo": False,
    }
    response = requests.post(
        WORKER_ADDR + "/worker_generate_completion",
        headers=headers,
        json=gen_params,
        stream=True,
    )
    return response

def initialize(model_nm, controller_addr=CONTROLLER_ADDRESS):
    global MODEL_NAME, WORKER_ADDR, CONTROLLER_ADDRESS
    MODEL_NAME = model_nm
    CONTROLLER_ADDRESS = controller_addr
    ret = requests.post(controller_addr + "/refresh_all_workers")
    ret = requests.post(controller_addr + "/list_models")
    models = ret.json()["models"]
    models.sort()


    ret = requests.post(
        controller_addr + "/get_worker_address", json={"model": model_nm}
    )
    WORKER_ADDR = ret.json()["address"]

    if WORKER_ADDR == "":
        return

def send_w_stream(messages, temperature=0.0, max_tokens= 500):
    global MODEL_NAME, WORKER_ADDR, CONTROLLER_ADDRESS
    ###
    ### build prompt
    #
    #conv = get_conversation_template(MODEL_NAME)
    #conv.append_message(conv.roles[0], args.message)
    #conv.append_message(conv.roles[1], None)
    #prompt = conv.get_prompt()

    ###
    ### build prompt
    #
    conv=fastchat.conversation.get_conv_template('wizard')
    """
    conv = Conversation(
        name="wizard",
        system="",
        roles=("USER", "ASSISTANT"),
        messages=[],
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
    """
    for msg in messages:
        role = msg['role']
        if role.lower() == 'user':
            role_index = 0
        else:
            role_index = 1
        conv.append_message(conv.roles[role_index], msg['content'])
    conv.append_message(conv.roles[1], '')
    prompt = conv.get_prompt()

    headers = {"User-Agent": "FastChat Client"}
    gen_params = {
        "model": MODEL_NAME,
        "prompt": prompt,
        "temperature": temperature,
        "max_new_tokens": max_tokens,
        "stop": conv.stop_str,
        "stop_token_ids": conv.stop_token_ids,
        "echo": False,
    }
    response = requests.post(
        WORKER_ADDR + "/worker_generate_stream",
        headers=headers,
        json=gen_params,
        stream=True,
    )

    prev = 0
    for chunk in response.iter_lines(decode_unicode=False, delimiter=b"\0"):
        if chunk:
            data = json.loads(chunk.decode())
            output = data["text"].strip()
            print(output[prev:], end="", flush=True)
            prev = len(output)
    print("")
    return output

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument("--worker-address", type=str)
    parser.add_argument("--model-name", type=str, required=True)
    parser.add_argument("--temperature", type=float, default=0.0)
    parser.add_argument("--max-new-tokens", type=int, default=32)
    parser.add_argument(
        "--message", type=str, default="Tell me a story with more than 1000 words."
    )
    args = parser.parse_args()

    main()



class MemoryFork:
    def __init__(self, memory):
        self._fork = dict()
        self._memory = memory

    def has(self, key):
        return key in self._fork or self._memory.has(key)

    def get(self, key):
        if key in self._fork:
            return self._fork[key]
        else:
            return self._memory.get(key)

    def set(self, key, value):
        self._fork[key] = value

    def delete(self, key):
        if key in self._fork:
            self._fork.pop(key, None)

    def clear(self):
        self._fork.clear()



import unittest
import yaml
from promptrix.VolatileMemory import VolatileMemory
from promptrix.FunctionRegistry import FunctionRegistry
from promptrix.GPT3Tokenizer import  GPT3Tokenizer
from promptrix.Prompt import Prompt

from alphawave.alphawaveTypes import PromptCompletionOptions
from alphawave.TestClient import TestClient
import asyncio

class TestClientTest(unittest.TestCase):
    def setUp(self):
        self.memory = VolatileMemory()
        self.functions = FunctionRegistry()
        self.tokenizer = GPT3Tokenizer()
        self.prompt = Prompt([])
        self.options = PromptCompletionOptions(
            completion_type='text',
            model='davinci',
        )

    def test_constructor_default_params(self):
        client = TestClient()
        self.assertEqual(client.status, 'success')
        self.assertEqual(client.response, {'role': 'assistant', 'content': 'Hello World'})

    def test_constructor_custom_params(self):
        client = TestClient('error', 'Hello Error')
        self.assertEqual(client.status, 'error')
        self.assertEqual(client.response, 'Hello Error')

    def test_completePrompt(self):
        client = TestClient()
        response = asyncio.run(client.completePrompt(self.memory, self.functions, self.tokenizer, self.prompt, self.options))
        self.assertEqual(response['status'], 'success')
        self.assertEqual(response['message'], {'role': 'assistant', 'content': 'Hello World'})

if __name__ == '__main__':
    unittest.main()



import requests
from dataclasses import dataclass, asdict
from typing import Optional, Dict, Any, Union
import time
import json
import asyncio
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, PromptSection, Tokenizer
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.AssistantMessage import AssistantMessage

from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse
from alphawave.internalTypes import ChatCompletionRequestMessage, CreateChatCompletionRequest, CreateChatCompletionResponse, CreateCompletionRequest, CreateCompletionResponse
from alphawave.Colorize import Colorize
import alphawave.utilityV2 as ut
import alphawave.LLMClient as client

@dataclass
class OSClientOptions(PromptCompletionOptions):
    def __init__(self, apiKey, organization = None, endpoint = None, logRequests = None):
        self.apiKey = apiKey
        self.organization = organization
        self.endpoint = endpoint
        self.logRequests = logRequests

def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if hasattr(instance, key):
            setattr(instance, key, value)

def get_values(instance, keys):
    values = []
    for key in keys:
        if hasattr(instance, key):
            values.append(getattr(instance, key))
        else:
            values.append(None)
    return values

        
@dataclass
class Response:
    status_code: int
    text: str
    headers: Dict[str,str] = None
    reason: str = ''
    
class OSClient(PromptCompletionClient):
    DefaultEndpoint = 'https://api.openai.com'
    UserAgent = 'AlphaWave'

    def __init__(self, **kwargs):
        self.options = OSClientOptions(apiKey=None, organization=None, endpoint= '127.0.0.1', logRequests=False)
        update_dataclass(self.options, **kwargs)
        if self.options.endpoint:
            self.options.endpoint = self.options.endpoint.strip()
            if self.options.endpoint.endswith('/'):
                self.options.endpoint = self.options.endpoint[:-1]

        self._session = requests.Session()

    async def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: PromptCompletionOptions) -> PromptResponse:
        if isinstance(options, dict):
            argoptions = options
            options = PromptCompletionOptions(completion_type = argoptions['completion_type'], model = argoptions['model'])
        startTime = time.time()
        max_input_tokens = 2048
        if hasattr(options, 'max_input_tokens') and getattr(options, 'max_input_tokens') is not None:
            max_input_tokens = options.max_input_tokens
        if hasattr(options, 'completion_type') and options.completion_type == 'text':
            result = prompt.renderAsText(memory, functions, tokenizer, max_input_tokens)
        if hasattr(options, 'completion_type') and options.completion_type == 'text':
            result = prompt.renderAsText(memory, functions, tokenizer, max_input_tokens)
            if result.tooLong:
                return {'status': 'too_long', 'message': f"The generated text completion prompt had a length of {result.length} tokens which exceeded the max_input_tokens of {max_input_tokens}."}
            if self.options.logRequests:
                print(Colorize.title('PROMPT:'))
                for msg in result.output:
                    if not isinstance(msg, dict):
                        msg = msg.__dict__
                    print(Colorize.output(json.dumps(msg, indent=2)), end='')
                print()
            request = self.copyOptionsToRequest(CreateCompletionRequest({
                'model': optionsmodel,
                'prompt': result.output,
            }), options, ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])
            response = self.createCompletion(request)
            if self.options.logRequests:
                print(Colorize.title('RESPONSE:'))
                print(Colorize.value('statuse', response.status))
                print(Colorize.value('duration', time.time() - startTime, 'ms'))
                print(Colorize.output(response.message))

            if response.status == 'success':
                completion = response.message
                return {'status': 'success', 'message': completion}
            else:
                return {'status': 'error', 'message': f"The text completion API returned an error status of {response.status}: {response.message}"}
        else:
            result = await prompt.renderAsMessages(memory, functions, tokenizer, max_input_tokens)
            if result.tooLong:
                return {'status': 'too_long', 'message': f"The generated chat completion prompt had a length of {result.length} tokens which exceeded the max_input_tokens of {max_input_tokens}."}
            self.options.logRequests = True
            if self.options.logRequests:
                print(Colorize.title('CHAT PROMPT:'))
                for msg in result.output:
                    if not isinstance(msg, dict):
                        msg = msg.__dict__
                    print(Colorize.output(json.dumps(msg, indent=2)), end='')
                print()
            request = self.copyOptionsToRequest(CreateChatCompletionRequest(model = options.model, messages =  result.output), options, ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])
            response = self.createChatCompletion(request)
            self.options.logRequests = True
            if self.options.logRequests:
                print(Colorize.title('CHAT RESPONSE:'))
                print(Colorize.value('status', response.status))
                print(Colorize.value('duration', time.time() - startTime, 'ms'))
                print(Colorize.output(response))

            if response.status == 'success':
                completion = response.message
                return {'status': 'success', 'message': completion}
            else:
                return {'status': 'error', 'message': f"The chat completion API returned an error status of {response.status}: {response.message}"}

    def addRequestHeaders(self, headers: Dict[str, str], options: OSClientOptions):
        headers['Authorization'] = f"Bearer {options.apiKey}"
        if options.organization:
            headers['OS-Organization'] = options.organization

    def copyOptionsToRequest(self, target: Dict[str, Any], src: Any, fields: list) -> Dict[str, Any]:
        for field in fields:
            if hasattr(src, field) and getattr(src, field) is not None:
                setattr(target,field, getattr(src,field))
        return target

    def createCompletion(self, request: CreateCompletionRequest) -> requests.Response:
        url = f"{self.options.endpoint or self.DefaultEndpoint}/v1/completions"
        return self.post(url, request)

    def createChatCompletion(self, request: CreateChatCompletionRequest) -> requests.Response:
        url = f"{self.options.endpoint or self.DefaultEndpoint}/v1/chat/completions"
        return self.post(url, request)

    def post(self, url: str, body: object) -> requests.Response:
        requestHeaders = {
            'Content-Type': 'application/json',
            'User-Agent': self.UserAgent
        }
        result = ''
        try:
            result = ut.ask_LLM(ut.MODEL, body.messages)
            runon_idx = result.find(client.USER)
            if runon_idx > 0:
                result = result[:runon_idx]
        except Exception as e:
            return PromptResponse(status='error',message=str(e))
        return PromptResponse(status='success', message = {'role':'assistant', 'content': result})



from pyee import AsyncIOEventEmitter
from typing import Dict, Any, Union
from alphawave_agents.agentTypes import Command, AgentThought, AgentThoughtSchema
from alphawave.JSONResponseValidator import JSONResponseValidator
import traceback

class AgentCommandValidator:
    def __init__(self, commands: dict[str, Command]):
        self._thought_validator = JSONResponseValidator(AgentThoughtSchema, 'No valid JSON objects were found in the response. Return a valid JSON object with your thoughts and the next command to perform.')
        self._commands = commands

    async def validate_response(self, memory, functions, tokenizer, response, remaining_attempts) -> Union[AgentThought, None]:
        # Validate that the response contains a thought
        try:
          validation_result = self._thought_validator.validate_response(memory, functions, tokenizer, response, remaining_attempts)
          if not validation_result['valid']:
              return validation_result

          # Validate that the command exists
          thought = validation_result['value']
          if not('command' in thought) or not('input' in thought['command']) or not ('name' in thought['command']):
              pass
          if not('command' in thought) or not('name' in thought['command']):
              pass

          command_name = thought['command']['input']['name'] if thought['command']['name'] == 'character' else thought['command']['name']
          #command_name = thought['command']['name']
          if command_name not in self._commands:
              return {
                  'type': 'Validation',
                  'valid': False,
                  'feedback': 'The command '+thought['command']['name']+f' does not exist. The only commands you have are {self._commands.keys()}'
              }
          
          # Validate that the command input is valid
          command = self._commands[command_name]
          command_validation_result = await command.validate(thought['command']['input'] or {}, memory, functions, tokenizer)
          if not command_validation_result['valid']:
              return command_validation_result
          
          # Return the validated thought
          return validation_result
        except Exception as e:
            pass
        return {
            'type': 'Validation',
            'valid': False,
            'feedback': 'The command '+thought['command']['name']+' validation failed, try again'
        }



import asyncio
import aiounittest, unittest
from assertpy import assert_that
from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, Tokenizer
from promptrix.FunctionRegistry import FunctionRegistry 
from promptrix.Prompt import Prompt
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.VolatileMemory import VolatileMemory
from alphawave.alphawaveTypes import PromptCompletionOptions, PromptResponse, PromptResponseValidator, Validation
from alphawave.DefaultResponseValidator import DefaultResponseValidator
from alphawave.TestClient import TestClient
from alphawave.OpenAIClient import OpenAIClient
from alphawave.AlphaWave import AlphaWave
import os
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.UserMessage import UserMessage
from promptrix.AssistantMessage import AssistantMessage


class TestValidator(PromptResponseValidator):
    def __init__(self, client):
        self.feedback = 'Something is wrong'
        self.repairAttempts = 0
        self.exception = None
        self.clientErrorDuringRepair = False
        self.returnContent = False
        self.client = client

    async def validateResponse(self, memory, functions, tokenizer, response, remaining_attempts):
        if self.exception:
            exception = self.exception
            self.exception = None
            raise exception
        print('Validator', self.clientErrorDuringRepair, self.repairAttempts)
        if self.clientErrorDuringRepair and self.repairAttempts == 0:
            self.clientErrorDuringRepair = False
            self.client.status = 'error'
            self.client.response = 'Some Error'
            return { 'type': 'Validation', 'valid': False, 'feedback': self.feedback }
        elif self.repairAttempts > 0:
            self.repairAttempts -= 1
            return { 'type': 'Validation', 'valid': False, 'feedback': self.feedback }
        elif self.returnContent:
            self.returnContent = False
            return { 'type': 'Validation', 'valid': True, 'value': response['message']['content'] }
        else:
            return { 'type': 'Validation', 'valid': True }

class TestAlphaWave(aiounittest.AsyncTestCase):
    def setUp(self):
        self.client = OpenAIClient(apiKey=os.getenv("OPENAI_API_KEY"))
        self.prompt = Prompt([
            SystemMessage('You are helpful, creative, clever, and very friendly.'),
            ConversationHistory('history', .5),
            UserMessage('{{$input}}', 100)
        ])

        self.prompt_options = PromptCompletionOptions(completion_type='chat', model='gpt-3.5-turbo')
        self.memory = VolatileMemory()
        self.memory.set('history', [])
        self.functions = FunctionRegistry()
        self.tokenizer = GPT3Tokenizer()
        self.validator = DefaultResponseValidator()

    def test_constructor(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator, history_variable='history', input_variable='input', max_repair_attempts=5, max_history_messages=20)
        assert_that(wave).is_not_none()
        assert_that(wave.options).is_not_none()
        assert_that(wave.options.client).is_equal_to(self.client)
        assert_that(wave.options.prompt).is_equal_to(self.prompt)
        assert_that(wave.options.prompt_options).is_equal_to(self.prompt_options)
        assert_that(wave.options.memory).is_equal_to(self.memory)
        assert_that(wave.options.functions).is_equal_to(self.functions)
        assert_that(wave.options.tokenizer).is_equal_to(self.tokenizer)
        assert_that(wave.options.validator).is_equal_to(self.validator)
        assert_that(wave.options.history_variable).is_equal_to('history')
        assert_that(wave.options.input_variable).is_equal_to('input')
        assert_that(wave.options.max_repair_attempts).is_equal_to(5)
        assert_that(wave.options.max_history_messages).is_equal_to(20)
    async def test_basic_prompt_completion(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        response = await wave.completePrompt()
        print(response['status'])
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'assistant', 'content':"))
        input = self.memory.get('input')
        assert_that(input).is_none()
        self.memory.clear()
        self.memory.set('history', [])
        

        self.client.response = 'Hello'
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content':"))
        input = self.memory.get('input')
        assert_that(input).is_equal_to('Hi')
        self.memory.clear()
        self.memory.set('history', [])
    """

    async def test_prompt_completion_with_validation(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.response = 'Hello'
        self.validator.repairAttempts = 1
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content':"))
        self.memory.clear()
        self.memory.set('history', [])

        self.client.response = 'Hello'
        self.validator.repairAttempts = 2
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content'"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content':"))
        self.memory.clear()

    async def test_prompt_completion_with_repair(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)

        print("starting test 1")
        
        self.client.response = 'Hello'
        self.validator.repairAttempts = 3
        response = await wave.completePrompt('Hi')
        #assert_that(response['status']).is_equal_to('invalid_response')
        history = self.memory.get('history')
        #assert_that(history).is_equal_to([])
        self.memory.clear()
        self.memory.set('history', [])
        
        print("starting test 2")

        self.client.response = 'Hello'
        self.validator.repairAttempts = 2
        self.validator.clientErrorDuringRepair = True
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        #assert_that(response['message']).is_equal_to('Some Error')
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_default_feedback(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = 'Hello'
        self.validator.repairAttempts = 1
        self.validator.feedback = None
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content': "))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': "))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_undefined_response(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = None
        self.validator.repairAttempts = 1
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content': '' }"))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_message_object_response(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = { 'role': 'assistant', 'content': { 'foo': 'bar'} }
        self.validator.repairAttempts = 1
        self.validator.returnContent = True
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': {"))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_repaired_response(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = { 'role': 'assistant', 'content': 'Hello World' }
        self.validator.repairAttempts = 1
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': "))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_parsed_content_object(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = { 'role': 'assistant', 'content': { 'foo': 'bar'} }
        self.validator.repairAttempts = 1
        self.validator.returnContent = True
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content': "))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': {"))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_repaired_response_undefined(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = None
        self.validator.repairAttempts = 1
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{'role': 'assistant', 'content': "))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_repaired_response_message_object(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = { 'role': 'assistant', 'content': 'Hello World' }
        self.validator.repairAttempts = 1
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content':"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': "))
        self.memory.clear()
        self.memory.set('history', [])

    async def test_prompt_completion_with_repaired_response_parsed_content_object(self):
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.prompt_options, memory=self.memory, functions=self.functions, tokenizer=self.tokenizer, validator=self.validator)
        
        self.client.status = 'success'
        self.client.response = { 'role': 'assistant', 'content': { 'foo': 'bar'} }
        self.validator.repairAttempts = 1
        self.validator.returnContent = True
        response = await wave.completePrompt('Hi')
        assert_that(response['status']).is_equal_to('success')
        assert_that(str(response['message']).startswith("{ 'role': 'assistant', 'content': {"))
        history = self.memory.get('history')
        assert_that(str(history).startswith("[{ 'role': 'user', 'content': 'Hi' },{ 'role': 'assistant', 'content': {"))
        self.memory.clear()
        self.memory.set('history', [])
    """
if __name__ == '__main__':
    unittest.main()



from jsonschema import validate, ValidationError
from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, Tokenizer
from alphawave.alphawaveTypes import PromptResponse, Validation, PromptResponseValidator
from alphawave.Response import Response
from pyee import AsyncIOEventEmitter
import json
import ast
import traceback
import re

class JSONResponseValidator(PromptResponseValidator):
    def __init__(self, schema=None, missing_json_feedback='Invalid JSON. Revise your previous response and return valid JSON as per the earlier schema.'):
        self.schema = schema
        self.missing_json_feedback = missing_json_feedback

    def parse_dict(self, s):
        if s is None:
            return s
        # Try to parse as JSON
        try:
            return json.loads(s)
        except json.JSONDecodeError:
            pass
        
        # Try to parse as a Python literal
        try:
            return ast.literal_eval(s)
        except (SyntaxError, ValueError):
            pass

        # Try to repair common errors and parse again
        s = s.strip()
        if not (s.startswith('{') and s.endswith('}')):
            s = '{' + s + '}'
        s = re.sub(r'([{,]\s*)(\w+)(\s*:)', r'\1"\2"\3', s)  # add quotes around keys
        #s = s.replace("'", '"')  # replace single quotes with double quotes

        # Try to parse the repaired string
        try:
            y = json.loads(s)
            return y
        except json.JSONDecodeError:
            return s

    def validate_response(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, response: PromptResponse, remaining_attempts: int) -> Validation:
        message = response['message']
        
        text = message if isinstance(message, str) else message.get('content', '')

        # Parse the response text
        parsed=[]
        try:
            parsed = Response.parse_all_objects(text)
        except Exception as e:
            raise e
        if len(parsed) == 0:
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': self.missing_json_feedback
            }

        # Validate the response against the schema
        if self.schema:
            errors = None
            for i in range(len(parsed) - 1, -1, -1):
                obj = parsed[i]
                try:
                    try:
                        obj = self.parse_dict(obj) if type(obj) == str else obj
                    except Exception as e:
                        pass
                    validate(obj, self.schema)
                    return {
                        'type': 'Validation',
                        'valid': True,
                        'value': obj
                    }
                except ValidationError as e:
                    path = str(list(e.relative_schema_path)[1:-1]).replace('[','').replace(']',"").replace(', ', ':')
                    if not errors:
                        errors = e
                    return {
                        'type': 'Validation',
                        'valid': False,
                        'feedback': f'The JSON returned had errors. Apply these fixes:\n{self.get_error_fix(errors)}. Revise your previous response to strict JSON format standards and this schema {self.schema}'
                    }
                except Exception as e:
                    return {
                        'type': 'Validation',
                        'valid': False,
                        'feedback': f'The JSON returned had errors. Apply these fixes:\n{self.get_error_fix(e)}. Revise your previous response to strict JSON format standards and this schema {self.schema}'
                    }      
    
        else:
            # Return the last object
            return {
                'type': 'Validation',
                'valid': True,
                'value': parsed[-1]
            }

    def get_error_fix(self, error: ValidationError) -> str:
        # Get argument as a string
        arg = error.validator
        path = str(list(error.relative_schema_path)[1:-1]).replace('[','').replace(']',"").replace(', ', ':')
        
        switcher = {
            'type': f'convert "{path}" value to a {error.validator_value}' if len(path)> 0 else '',
            'anyOf': f'convert "{path}" to one of the allowed types: {error.validator_value}',
            'additionalProperties': f'remove the "{arg}" field from "{path}"',
            'required': f"add the {error.validator_value} fields to {path if len(path)>0 else 'response'}",
            'format': f'change the "{path}" field to be a {error.validator_value}',
            'uniqueItems': f'remove all duplicate items from "{error.path}"',
            'enum': f'change the "{path}" value to be one of these values: {arg}',
            'const': f'change the "{path}" value to be {arg}',
        }

        return switcher[arg] if arg in switcher else error.message




import json
import re
class Response:
    @staticmethod
    def parse_all_objects(text):
        objects = []
        if text is None:
            return objects
        lines = text.split('\n')
        if len(lines) > 1:
            for line in lines:
                obj = Response.parse_json(line)
                if obj:
                    objects.append(obj)

        if len(objects) == 0:
            obj = Response.parse_json(text)
            if obj:
                objects.append(obj)

        #print(f'***** Response return \n{objects}\n')
        return objects

    @staticmethod
    def parse_json(text):
        text = ''.join(c for c in text if c.isprintable())
        text = text.replace('{\n', '{')
        text = text.replace('}\n', '}')
        #text = re.sub(r"'([^\"']+)'", r'"\1"', text) # all pairs as doublequote
        text = re.sub(r"'([^\"']+)':", r'"\1":', text) # keys as doublequote
        #text = re.sub(r'"([^\'"]+)":', r"'\1':", text) # keys as singlequote
        #text = text.replace("'", '"')
        #text = text.replace("\'", '"')
        start_brace = text.find('{')
        if start_brace >= 0:
            obj_text = text[start_brace:]
            nesting = ['}']
            cleaned = '{'
            in_string = False
            i = 1
            while i < len(obj_text) and len(nesting) > 0:
                ch = obj_text[i]
                if in_string:
                    cleaned += ch
                    if ch == '\\':
                        i += 1
                        if i < len(obj_text):
                            cleaned += obj_text[i]
                        else:
                            return None
                    elif ch == '"':
                        in_string = False
                else:
                    if ch == '"':
                        in_string = True
                    elif ch == '{':
                        nesting.append('}')
                    elif ch == '[':
                        nesting.append(']')
                    elif ch == '}':
                        close_object = nesting.pop()
                        if close_object != '}':
                            return None
                    elif ch == ']':
                        close_array = nesting.pop()
                        if close_array != ']':
                            return None
                    elif ch == '<':
                        ch = '"<'
                    elif ch == '>':
                        ch = '>"'
                    cleaned += ch
                i += 1

            if len(nesting) > 0:
                cleaned += ''.join(reversed(nesting))

            try:
                if type(cleaned) == str:
                    obj = json.loads(cleaned)
                    return obj
                else:
                    return cleaned
                return obj if len(obj.keys()) > 0 else None
            except json.JSONDecodeError:
                return cleaned
        else:
            return None



from promptrix.promptrixTypes import PromptFunctions, PromptMemory, Tokenizer
from alphawave.alphawaveTypes import PromptResponse, Validation, PromptResponseValidator

class DefaultResponseValidator(PromptResponseValidator):
    def validate_response(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, response: PromptResponse, remaining_attempts) -> Validation:
        self.feedback = response['message']['content'] if isinstance(response['message'], dict) else response.message
        return {
            'type': 'Validation',
            'valid': True,
            'value': response['message']['content'] if isinstance(response['message'], dict) else response.message
        }



from typing import Dict, Optional
from promptrix.promptrixTypes import PromptMemory, PromptFunctions, Tokenizer
from promptrix.Prompt import Prompt
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand
from alphawave_agents.SchemaBasedCommand import CommandSchema as sbcCommandSchema
from alphawave_agents.agentTypes import TaskResponse
from dataclasses import dataclass, asdict

@dataclass
class CommandSchema(sbcCommandSchema):
    schema_type: str
    title: str
    description: str
    properties: Dict[str,Dict[str,str]]
    required: list[str]
    returns: str

schema = CommandSchema(
    schema_type="object",
    title= "ask",
    description= "ask the user a question and wait for their response",
    properties= {
        'question': {
            'type': "string",
            'description': "question to ask"
        }
    },
    required= ["question"],
    returns= "users answer"
)


class AskCommandInput:
    def __init__(self, question: str):
        self.question = question

class AskCommand(SchemaBasedCommand):
    def __init__(self, title: Optional[str] = '', description: Optional[str] = ''):
        global schema
        super().__init__(schema, title, description)

    def execute(self, input: AskCommandInput, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer) -> TaskResponse:
        return asdict(TaskResponse(status="input_needed",message=input['question']))



import requests, time, copy
from typing import Optional, Dict, Any, Union
from dataclasses import dataclass, asdict
import json
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, PromptSection, Tokenizer
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.AssistantMessage import AssistantMessage

from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse
from alphawave.internalTypes import ChatCompletionRequestMessage, CreateChatCompletionRequest, CreateChatCompletionResponse, CreateCompletionRequest, CreateCompletionResponse
from alphawave.Colorize import Colorize

@dataclass
class OpenAIClientOptions:
    def __init__(self, apiKey=None, organization = None, endpoint = None, logRequests = False):
        self.apiKey = apiKey
        self.organization = organization
        self.endpoint = endpoint
        self.logRequests = logRequests

def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if hasattr(instance, key):
            setattr(instance, key, value)

class OpenAIClient(PromptCompletionClient):
    DefaultEndpoint = 'https://api.openai.com'
    UserAgent = 'AlphaWave'

    def __init__(self, **kwargs):
        self.options = {'apiKey':None, 'organization':None, 'endpoint':None, 'logRequests':False}
        self.options.update(kwargs)
        if self.options['endpoint']:
            self.options['endpoint'] = self.options['endpoint'].strip()
            if self.options['endpoint'].endswith('/'):
                self.options['endpoint'] = self.options['endpoint'][:-1]

            if not self.options['endpoint'].lower().startswith('https://'):
                raise ValueError(f"Client created with an invalid endpoint of '{options['endpoint']}'. The endpoint must be a valid HTTPS url.")

        if not self.options['apiKey']:
            print("Client created without an 'apiKey'.")
            raise ValueError

        self._session = requests.Session()

    async def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: PromptCompletionOptions) -> PromptResponse:
        startTime = time.time()
        max_input_tokens = 1024
        if isinstance(options, dict):
            argoptions = options
            options = PromptCompletionOptions(completion_type = argoptions['completion_type'], model = argoptions['model'])
            update_dataclass(options, **argoptions)
        if hasattr(options, 'max_input_tokens') and getattr(options, 'max_input_tokens') is not None:
            max_input_tokens = options.max_input_tokens
        if hasattr(options, 'completion_type') and options.completion_type == 'text':
            result = prompt.renderAsText(memory, functions, tokenizer, max_input_tokens)
            if result.tooLong:
                return {'status': 'too_long', 'message': f"The generated text completion prompt had a length of {result.length} tokens which exceeded the max_input_tokens of {max_input_tokens}."}
            if self.options['logRequests']:
                print(Colorize.title('PROMPT:'))
                for msg in result.output:
                    if not isinstance(msg, dict):
                        msg = msg.__dict__
                    print(Colorize.output(json.dumps(msg, indent=2)), end='')
                print()

            request = self.copyOptionsToRequest(CreateCompletionRequest({
                'model': self.options['model'],
                'prompt': result['output'],
            }), options, ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])
            response = self.createCompletion(request)
            if self.options['logRequests']:
                print(Colorize.title('RESPONSE:'))
                print(Colorize.value('status', response.status))
                print(Colorize.value('duration', time.time() - startTime, 'ms'))
                print(Colorize.output(response.json()))

            if response.status_code < 300:
                completion = response.json().get('choices')[0]
                return {'status': 'success', 'message': {'role': 'assistant', 'content': completion.get('text', '')}}
            elif response.status_code == 429:
                if self.options['logRequests']:
                    print(Colorize.title('HEADERS:'))
                    print(Colorize.output(response.headers))
                return {'status': 'rate_limited', 'message': 'The text completion API returned a rate limit error.'}
            else:
                return {'status': 'error', 'message': f"The text completion API returned an error status of {response.status_code}: {response.reason}"}
        else:
            result = await prompt.renderAsMessages(memory, functions, tokenizer, max_input_tokens)
            if result.tooLong:
                return {'status': 'too_long', 'message': f"The generated chat completion prompt had a length of {result.length} tokens which exceeded the max_input_tokens of {max_input_tokens}."}
            if self.options['logRequests']:
                print(Colorize.title('CHAT PROMPT:'))
                for msg in result.output:
                    if not isinstance(msg, dict):
                        msg = msg.__dict__
                    print(Colorize.output(json.dumps(msg, indent=2)), end='')
                print()

            request = self.copyOptionsToRequest(CreateChatCompletionRequest(model=options.model, messages=result.output), options,
                                                    ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])
            response = self.createChatCompletion(request)
            if self.options['logRequests']:
                print(Colorize.title('CHAT RESPONSE:'))
                print(Colorize.value('statuse', response.status_code))
                print(Colorize.value('duration', time.time() - startTime, 'ms'))
                print(Colorize.output(response.json()))

            if response.status_code < 300:
                completion = response.json().get('choices')[0]
                return {'status': 'success', 'message': completion.get('message', {'role': 'assistant', 'content': ''})}
            elif response.status_code == 429:
                if self.options['logRequests']:
                    print(Colorize.title('HEADERS:'))
                    print(Colorize.output(response.headers))
                return {'status': 'rate_limited', 'message': 'The chat completion API returned a rate limit error.'}
            else:
                return {'status': 'error', 'message': f"The chat completion API returned an error status of {response.status_code}: {response.reason}"}

    def addRequestHeaders(self, headers: Dict[str, str], options: OpenAIClientOptions):
        headers['Authorization'] = f"Bearer {options['apiKey']}"
        if options['organization']:
            headers['OpenAI-Organization'] = options['organization']

    def copyOptionsToRequest(self, target: Dict[str, Any], src: Any, fields: list) -> Dict[str, Any]:
        for field in fields:
            if hasattr(src, field) and getattr(src, field) is not None:
                setattr(target,field, getattr(src,field))
        return target

    def createCompletion(self, request: CreateCompletionRequest) -> requests.Response:
        url = f"{self.options['endpoint'] or self.DefaultEndpoint}/v1/completions"
        return self.post(url, request)

    def createChatCompletion(self, request: CreateChatCompletionRequest) -> requests.Response:
        url = f"{self.options['endpoint'] or self.DefaultEndpoint}/v1/chat/completions"
        return self.post(url, request)

    def post(self, url: str, body: object) -> requests.Response:
        requestHeaders = {
            'Content-Type': 'application/json',
            'User-Agent': self.UserAgent
        }
        self.addRequestHeaders(requestHeaders, self.options)
        jsonbody = asdict(body)
        keys = list(jsonbody.keys())
        for key in keys:
            if jsonbody[key] is None:
                del jsonbody[key]
        result = self._session.post(url, json=jsonbody, headers=requestHeaders)
        if result.status_code < 300:
            completion = result.json().get('choices')[0]

        return result
        



from typing import Dict, Optional
from abc import ABC, abstractmethod

class CommandSchema:
    def __init__(self, type: str, title: str, description: str, properties: Dict, required: list):
        self.type = type
        self.title = title
        self.description = description
        self.properties = properties
        self.required = required

class SchemaBasedCommand(ABC):
    def __init__(self, schema: CommandSchema, title: Optional[str] = None, description: Optional[str] = None):
        self.schema = schema
        self.title = title
        self.description = description

    @abstractmethod
    def execute(self, input, memory, functions, tokenizer):
        pass

class CompleteTaskCommandInput:
    def __init__(self, status: str):
        self.status = status

class TaskResponse:
    def __init__(self, type: str, status: str, message: str):
        self.type = type
        self.status = status
        self.message = message

class CompleteTaskCommand(SchemaBasedCommand):
    def __init__(self, response: Optiona[str] = None, title: Optional[str] = None, description: Optional[str] = None):
        schema = CommandSchema(
            type="object",
            title="completeTask",
            description="signals that the task is completed",
            properties={
                "status": {
                    "type": "string",
                    "description": "brief completion status"
                }
            },
            required=["status"]
        )
        super().__init__(schema, title, description)

    def execute(self, input: CompleteTaskCommandInput, memory, functions, tokenizer):
        rsp = input.status
        if self.response is not None:
            rsp = self.response
        return asdict(TaskResponse(
            type="TaskResponse",
            status="success",
            message=rsp
        ))



from typing import Callable, Any, Dict, Union
from dataclasses import dataclass, asdict, field
from promptrix.promptrixTypes import PromptMemory, PromptFunctions, Tokenizer
from promptrix.Utilities import Utilities
from promptrix.Prompt import Prompt
from alphawave.AlphaWave import AlphaWave
from alphawave.AlphaWave import AlphaWaveOptions
from alphawave.alphawaveTypes import PromptCompletionOptions
from alphawave.MemoryFork import MemoryFork
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand, CommandSchema as sbcCommandSchema
from alphawave_agents.agentTypes import TaskResponse

@dataclass
class CommandSchema(sbcCommandSchema):
    schema_type: str
    title: str
    description: str
    properties: Dict[str,Dict[str,str]]
    required: list[str] = field(default_factory=list)
    returns: str = None

def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if hasattr(instance, key):
            setattr(instance, key, value)

    
@dataclass
class PromptCommandOptions(AlphaWaveOptions):
    def __init__(self, prompt_options: PromptCompletionOptions, schema: CommandSchema, parseResponse: Callable[[str, Dict[str, Any], PromptMemory, PromptFunctions, Tokenizer], Any] = None):
        super().__init__(prompt_options)
        self.prompt_options=prompt_options
        self.schema=schema
        self.parseResponse=parseResponse

class PromptCommand(SchemaBasedCommand):
    def __init__(self, client, prompt: Prompt, options: PromptCommandOptions, title: str = None, description: str = None):
        super().__init__(options.schema, title, description)
        self.options = options
        self.client = client
        self.prompt = prompt
        self.prompt_options = options.prompt_options

    async def execute(self, input: Dict[str, Any], memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer) -> Union[TaskResponse, str]:
        # Fork memory and copy the input into the fork
        fork = MemoryFork(memory)
        for key, value in input.items():
            fork.set(key, value)

        # Create a wave and send it
        #options = AlphaWaveOptions()
        #update_dataclass(options, **self.options.__dict__)
        #update_dataclass(options, memory=fork, functions= functions, tokenizer= tokenizer)
        wave = AlphaWave(client=self.client, prompt=self.prompt, prompt_options=self.options.prompt_options, memory=memory, functions=functions, tokenizer=tokenizer)
        response = await wave.completePrompt()
        # Process the response
        message = response['message']['content'] if isinstance(response['message'], dict) else response['message']
        if response['status'] == "success":
            # Return the response
            parsed = await self.options.parseResponse(message, input, memory, functions, tokenizer) if self.options.parseResponse else message
            return Utilities.to_string(tokenizer, parsed)
        else:
            # Return the error
            return {
                "type": "TaskResponse",
                "status": response['status'],
                "message": message
            }




from pyee import AsyncIOEventEmitter
from jsonschema import validate, ValidationError
from typing import Any, Dict, Optional, Union
#import json
from promptrix.promptrixTypes import PromptMemory, PromptFunctions, Tokenizer
from dataclasses import dataclass, asdict
import traceback

@dataclass
class CommandSchema:
    def __init__(self, type: str, title: str='', description: str='', returns: Optional[str] = None):
        self.type = type
        self.title = title
        self.description = description
        self.returns = returns

class SchemaBasedCommand(AsyncIOEventEmitter):
    def __init__(self, schema: CommandSchema, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__()
        self._schema = schema
        self._title = title
        self._description = description

    @property
    def description(self) -> str:
        return self._description or self._schema.description

    @property
    def inputs(self) -> Optional[str]:
        if self._schema.properties:
            properties = self._schema.properties or {}
            inputs = [f'"{key}":"<{property.get("description", property.get("type", "any"))} value>"' for key, property in properties.items()]
            return ",".join(inputs)
        else:
            return None

    @property
    def output(self) -> Optional[str]:
        return self._schema.returns

    @property
    def schema(self) -> CommandSchema:
        return self._schema

    @property
    def title(self) -> str:
        return self._title or self._schema.title

    async def execute(self, input: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer') -> Any:
        raise NotImplementedError

    async def validate(self, input: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer') -> 'Validation':
        try:
            cleaned = self.clean_input(input)
            if self._schema is not None:
                validate(cleaned, asdict(self._schema))
            return {
                'type': 'Validation',
                'valid': True,
                'value': cleaned
            }
        except ValidationError as e:
            errors = f'"{e.path[0] if e.path else "input"}": {e.message}'
            message = errors
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': f"The ['command']['input'] field has errors:\n{message}\n\nTry again."
            }

    def clean_input(self, input: Dict[str, Any]) -> Dict[str, Any]:
        cleaned = {}
        properties = self._schema.properties
        if not properties:
            return cleaned

        for key, property in properties.items():
            type = property.get('type', 'any')
            value = input.get(key)

            if value is None or (isinstance(value, str) and value.startswith("<") and value.endswith(">")):
                continue

            if not isinstance(value, str):
                value = json.dumps(value) if isinstance(value, dict) else str(value)

            try:
                if type == 'string':
                    cleaned[key] = str(value)
                elif type == 'number':
                    cleaned[key] = float(value)
                elif type == 'boolean':
                    cleaned[key] = bool(value)
                elif type in ['array', 'object']:
                    cleaned[key] = json.loads(value)
            except Exception:
                pass

        return cleaned
